/*
  Tool            Test_SearchPath
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE v4 upwards, all variants
  Version         v1.0.1   27 Mar 2024


  This tool tests TSE's SearchPath() statement.

  Its code contains a part dedicated to be copyable into sanity1.
  This part is clearly marked with comments.
  For the part to be usable in sanity1's context:
  - In its code all globally referenced objects have a name that starts with
    "test_SearchPath_" to avoid conflict with the rest of the sanity1 code.
  - The test_SearchPath() function returns an integer value TRUE or FALSE for
    success or failure, just like sanity1 expects.

  Note:
    TSE's SearchPath() command was changed in TSE 4.50rc20 (Internal version
    12386).
    The tool is compatible with both the old and the new SearchPath().
    For the new SearchPath() it performs each test twice; once for each value
    of SearchCurrentDirectory.

  Notable option:
    The global variable "test_SearchPath_debug" can be set to ON to generate
    an enlightening debug buffer as a side-effect.

  Inner workings:
    The tool generates a unique name, that is non-existent across all used
    directories.
    With that unique name it creates a work directory in the operating system's
    temporary directory, and in this work directory it creates a temporary
    current directory and three path directories to be used in tests.

    SearchPath() tests are defined as "test_SearchPath_test()" in a specific
    part of the test_SearchPath procedure that is clearly marked with comments.

    Each test_SearchPath_test() test actually does a lot of SearchPath() tests.
    SearchPath() is tested by comparing "all directories in which SearchPath()
    is expected to be able find the unique file in" against "all directories
    SearchPath() actually finds  the unique file in".
    The correct order of directories to find the file in is checked as well.
    And if the SearchCurrentDirectory setting is available, then each test
    is done twice, namely once for each of the setting's values.

    Each SearchPath() test must adhere to these restrictions:
    - Searches must be for the generated unique name in variable "unique_name".
    - SearchPath()'s 3rd parameter must be "mac", "." or "".

  Scope:
    Not tested for (yet?):
      SearchPath(filename, "")
        Undocumented.
        Probably searches the current directory and then the OS's path settings.
        The OS's path settings are not limited to TSE's 255 byte limit.

      SearchPath(filename, "", subdir)
         Undocumented and unknown.

    Tested for:
      SearchPath(FullyQualifiedFilename, IrrelevantParameter)
        Does not search.
        Just returns FullyQualifiedFilename if it exists.

      SearchPath(RootlessFilename, path)
        Only searches in path.

      SearchPath(RootlessFilename, path, subdir)
        Optionally searches the current directory:
          In TSE v4.50rc19 downwards this is always done.
          In TSE v4.50rc20 upwards   this is done optionally depending on the
          boolean editor variable SearchCurrentDirectory.
        Then searches each dir in the path, and immediately after each dir
        searches dir/subdir.
        Then searches LoadDir() and then LoadDir() + subdir.

      SearchPath(RootlessFilename, path, ".")
        Optionally searches the current directory:
          In TSE v4.50rc19 downwards this is always done.
          In TSE v4.50rc20 upwards   this is done optionally depending on the
          boolean editor variable SearchCurrentDirectory.
        Then searches each dir in the path.
        Then searches LoadDir().


  HISTORY

  v1       25 Mar 2024
    Initial release.

  v1.0.1   27 Mar 2024
    Bug fixed: Was not compatible with Semware's Sanity1 tool.

    Two commented meta-tests added to intentionally produce an error
    in two different forms.
*/





// Start of general compatibility restrictions and mitigations ...


#ifdef LINUX
  #define WIN32 FALSE
#endif


#ifndef WIN32
   16-bit versions of TSE are not supported. You need at least TSE 4.0.
#endif

#ifndef EDITOR_VERSION
   Editor Version is older than TSE 3.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4000h
   Editor Version is older than TSE 4.0. You need at least TSE 4.0.
#endif


#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif


#if INTERNAL_VERSION == 0
  #ifdef LINUX
    string proc GetTempPath()
      return('/tmp/')
    end GetTempPath
  #else
    string proc GetTempPath()
      return(AddTrailingSlash(GetEnvStr('TMP')))
    end GetTempPath
  #endif
#endif


#if EDITOR_VERSION < 4200h
  /*
    This procedure implements the MkDir() command of TSE 4.2 upwards.
  */
  integer proc MkDir(string dir)
    Dos('MkDir ' + QuotePath(dir), _DONT_PROMPT_|_DONT_CLEAR_|_START_HIDDEN_|_RETURN_CODE_)
    return(not DosIOResult())
  end MkDir
#endif


#if EDITOR_VERSION < 4200h
  integer proc RmDir(string dir)
    Dos('RmDir ' + QuotePath(dir), _DONT_PROMPT_|_DONT_CLEAR_|_START_HIDDEN_|_RETURN_CODE_)
    return(not DosIOResult())
  end RmDir
#endif


// End of general compatibility restrictions and mitigations.





// Start of test_SearchPath() implementation ...


string  test_SearchPath_cur_dir     [MAXSTRINGLEN] = ''
integer test_SearchPath_debug                      = OFF
integer test_SearchPath_debug_id                   = 0
integer test_SearchPath_id                         = 0
string  test_SearchPath_path1       [MAXSTRINGLEN] = ''
string  test_SearchPath_path2       [MAXSTRINGLEN] = ''
string  test_SearchPath_path3       [MAXSTRINGLEN] = ''
integer test_SearchPath_scd                        = 0
string  test_SearchPath_subdir      [MAXSTRINGLEN] = ''
string  test_SearchPath_unique_name [MAXSTRINGLEN] = ''


integer proc test_SearchPath_make_files(string dir)
  string  fqn [MAXSTRINGLEN] = ''
  integer ok                 = TRUE

  // This proc always returns TRUE.
  // SaveAs() errors are acceptable here.

  fqn = AddTrailingSlash(dir) + test_SearchPath_unique_name

  if not FileExists(fqn)
    SaveAs(fqn, _DONT_PROMPT_|_OVERWRITE_)
  endif

  fqn = AddTrailingSlash(AddTrailingSlash(dir) + test_SearchPath_subdir) +
                         test_SearchPath_unique_name

  if not FileExists(fqn)
    SaveAs(fqn, _DONT_PROMPT_|_OVERWRITE_)
  endif

  return(ok)
end test_SearchPath_make_files


integer proc test_SearchPath_delete_dirs(string dir)
  string  subdir_fqn [MAXSTRINGLEN] = AddTrailingSlash(dir) + test_SearchPath_subdir
  integer ok                        = TRUE

  if FileExists(dir) & _DIRECTORY_
    if FileExists(subdir_fqn) & _DIRECTORY_
      RmDir(subdir_fqn)
    endif

    RmDir(dir)
  endif

  return(ok)
end test_SearchPath_delete_dirs


integer proc test_SearchPath_delete_files(string dir)
  string  file_fqn        [MAXSTRINGLEN] = ''
  integer ok                             = TRUE
  string  subdir_file_fqn [MAXSTRINGLEN] = ''

  file_fqn        = AddTrailingSlash(dir) +
                    test_SearchPath_unique_name

  subdir_file_fqn = AddTrailingSlash(AddTrailingSlash(dir)   +
                                     test_SearchPath_subdir) +
                    test_SearchPath_unique_name

  if FileExists(file_fqn)
    EraseDiskFile(file_fqn)
  endif

  if FileExists(subdir_file_fqn)
    EraseDiskFile(subdir_file_fqn)
  endif

  return(ok)
end test_SearchPath_delete_files


integer proc test_SearchPath_expect_file(string dir)
  string  file_fqn [MAXSTRINGLEN] = AddTrailingSlash(dir) + test_SearchPath_unique_name
  integer ok                      = TRUE

  if test_SearchPath_debug
    AddLine('Expected: "' + file_fqn + '"', test_SearchPath_debug_id)
  endif

  if FileExists(file_fqn)
    if lFind(file_fqn, '^gi$')
      EndFile()
    else
      AddLine(file_fqn)
    endif
  endif

  return(ok)
end test_SearchPath_expect_file


integer proc test_SearchPath_expect_subdir_file(string dir)
  integer ok                        = TRUE
  string  subdir_fqn [MAXSTRINGLEN] = ''

  subdir_fqn = AddTrailingSlash(dir) + test_SearchPath_subdir
  ok         = test_SearchPath_expect_file(subdir_fqn)

  return(ok)
end test_SearchPath_expect_subdir_file


string proc test_SearchPath_option(string fn, string path, string subdir)
  integer counter                     = 0
  string  dir          [MAXSTRINGLEN] = ''
  string  expected_fqn [MAXSTRINGLEN] = ''
  string  found_fqn    [MAXSTRINGLEN] = ''
  integer is_fqn                      = FALSE
  string  result       [MAXSTRINGLEN] = ''

  if test_SearchPath_debug
    PushLocation()
    GotoBufferId(test_SearchPath_debug_id)

    AddLine('')
    AddLine('SearchPath("' + fn     + '",')
    AddLine('           "' + path   + '",')
    AddLine('           "' + subdir + '")')
    AddLine('      with SearchCurrentDirectory ' +
                   iif(test_SearchPath_scd, 'ON', 'OFF') + '.')
    PopLocation()
  endif

  if NumTokens(path, ';') == 0
    return('Testing "OS PATH search" with empty "path" parameter is not implemented.')
  endif

  // Make sure all searchable files exist irrespective of parameters and option.
  test_SearchPath_make_files(test_SearchPath_cur_dir)
  test_SearchPath_make_files(test_SearchPath_path1  )
  test_SearchPath_make_files(test_SearchPath_path2  )
  test_SearchPath_make_files(test_SearchPath_path3  )
  test_SearchPath_make_files(LoadDir()              )

  // Generate a list of expected files to be found in the expected order,
  // based on the search parameters and our own test_SearchPath_scd option.
  EmptyBuffer()

  if (WhichOS() in _WINDOWS_, _WINDOWSNT_)
    if (   isAlpha(fn[1])
       and fn[2] == ':')
    or fn[1:2] == '\\'
      is_fqn = TRUE
    endif
  else
    if fn[1] == '/'
      is_fqn = TRUE
    endif
  endif

  if is_fqn
    //  Searching a file by its fully qualified name should have only itself as
    //  a result, irrespective of other parameters.
    if FileExists(fn)
      AddLine(fn)   // Just add the fully qualified name as the expected file.

      if test_SearchPath_debug
        AddLine('Expected: "' + fn + '"', test_SearchPath_debug_id)
      endif
    endif
  else
    if  subdir <> ''
    and test_SearchPath_scd
      test_SearchPath_expect_file(test_SearchPath_cur_dir)
    endif

    //  Windows' PATH setting uses ";" as directory separator.
    //  Linux'   PATH setting uses ":" as directory separator.
    //  The SearchPath() statement of both Windows TSE and Linux TSE
    //  expects a ";" separator in its path parameter.
    for counter = 1 to NumTokens(path, ';')
      dir = GetToken(path, ';', counter)
      test_SearchPath_expect_file(dir)

      if not (subdir in '.', '')
        test_SearchPath_expect_subdir_file(dir)
      endif
    endfor

    if subdir <> ''
      test_SearchPath_expect_file(LoadDir())
      if subdir <> '.'
        test_SearchPath_expect_subdir_file(LoadDir())
      endif
    endif
  endif

  //  Now repeatedly use SearchPath() to find and delete each expected file in
  //  order.
  BegFile()
  repeat
    expected_fqn = GetText(1, MAXSTRINGLEN)

    if subdir == ''
      found_fqn = SearchPath(fn, path)
    else
      found_fqn = SearchPath(fn, path, subdir)
    endif

    if test_SearchPath_debug
      AddLine('Found   : "' + found_fqn + '"', test_SearchPath_debug_id)
    endif

    if EquiStr(expected_fqn, found_fqn)
      KillLine()
      EraseDiskFile(found_fqn)
    else
      PushLocation()
      test_SearchPath_debug = TRUE
      GotoBufferId(test_SearchPath_debug_id)
      AddLine('')
      AddLine('SearchPath() error!')
      AddLine('')
      AddLine('SearchPath("' + fn     + '",')
      AddLine('           "' + path   + '",')
      AddLine('           "' + subdir + '")')
      AddLine('with SearchCurrentDirectory == ' + iif(test_SearchPath_scd, 'ON', 'OFF'))
      AddLine('returns')
      AddLine('   "' + found_fqn    + '"')
      AddLine('instead of')
      AddLine('   "' + expected_fqn + '".')
      AddLine('')
      UpdateDisplay(_ALL_WINDOWS_REFRESH_)
      result = 'SearchPath() error: See description in new "' +
               CurrFilename() + '" buffer.'
      PopLocation()
    endif
  until result       <> ''
     or expected_fqn == ''
  return(result)
end test_SearchPath_option


string proc test_SearchPath_test(string fn, string path, string subdir)
  string result [MAXSTRINGLEN] = ''

  result = test_SearchPath_option(fn, path, subdir)

  #if INTERNAL_VERSION >= 12387
    if result == ''
      Toggle(searchcurrentdirectory)     // For SearchPath()'s internal use.
      test_SearchPath_scd = not test_SearchPath_scd // For this macro's use.

      result = test_SearchPath_option(fn, path, subdir)

      Toggle(searchcurrentdirectory)     // For SearchPath()'s internal use.
      test_SearchPath_scd = not test_SearchPath_scd // For this macro's use.
    endif
  #endif

  return(result)
end test_SearchPath_test


string proc test_SearchPath()
  string  cur_dir         [MAXSTRINGLEN] = ''
  string  result          [MAXSTRINGLEN] = ''
  string  old_current_dir [MAXSTRINGLEN] = CurrDir()
  string  path1           [MAXSTRINGLEN] = ''
  string  path2           [MAXSTRINGLEN] = ''
  string  path3           [MAXSTRINGLEN] = ''
  string  subdir          [MAXSTRINGLEN] = 'Mac'
  string  time                      [11] = Str(GetTime())
  string  tmp_dir         [MAXSTRINGLEN] = AddTrailingSlash(GetTempPath())
  integer tries                          = 0
  string  unique_name     [MAXSTRINGLEN] = ''
  string  wrk_dir         [MAXSTRINGLEN] = ''

  // Start of setting up the tests ...

  PushLocation()
  test_SearchPath_debug_id = NewFile()
  ChangeCurrFilename(SplitPath(CurrMacroFilename(), _NAME_) +
                     ':Debug:' + time,
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
  test_SearchPath_id       = CreateTempBuffer()
  ChangeCurrFilename(SplitPath(CurrMacroFilename(), _NAME_) +
                     ':Tmp:' + time,
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)

  //  test_SearchPath_id remains the current buffer until either all tests are
  //  done or an error is encountered.

  #if INTERNAL_VERSION < 12387
    test_SearchPath_scd = ON
  #else
    test_SearchPath_scd = Query(searchcurrentdirectory)
  #endif

  // Determine a unique name that does not exist in any fixed SearchPath
  // directory or the temporary directory.
  repeat
    tries = tries + 1
    if tries <= 10000
      unique_name = SplitPath(MakeTempName(LoadDir()), _NAME_|_EXT_)
    else
      result = 'Program error: No unique name.'
    endif
  until result <> ''
    or (    not FileExists(LoadDir()       + 'mac\' + unique_name)
        and not FileExists(old_current_dir          + unique_name)
        and not FileExists(old_current_dir + 'mac\' + unique_name)
        and not FileExists(tmp_dir                  + unique_name))

  // Create all possible non-fixed SearchPath directories.
  if      result == ''
  and not MkDir(tmp_dir + unique_name)
    result = 'Program error: MkDir ' + QuotePath(tmp_dir + unique_name)
  endif

  if result == ''
    // Create temporary helper directories
    wrk_dir = AddTrailingSlash(tmp_dir + unique_name)
    cur_dir = wrk_dir + 'CurrentDir'
    path1   = wrk_dir + 'Path1'
    path2   = wrk_dir + 'Path2'
    path3   = wrk_dir + 'Path3'
    MkDir(cur_dir)
    MkDir(path1)
    MkDir(path2)
    MkDir(path3)
    MkDir(AddTrailingSlash(cur_dir) + subdir)
    MkDir(AddTrailingSlash(path1)   + subdir)
    MkDir(AddTrailingSlash(path2)   + subdir)
    MkDir(AddTrailingSlash(path3)   + subdir)
  endif

  // Communicate static variables thru global variables to increase readability.
  test_SearchPath_cur_dir     = cur_dir
  test_SearchPath_path1       = path1
  test_SearchPath_path2       = path2
  test_SearchPath_path3       = path3
  test_SearchPath_subdir      = subdir
  test_SearchPath_unique_name = unique_name

  if      result == ''
  and     (WhichOS() in _WINDOWSNT_, _WINDOWS_)
  and not LogDrive(cur_dir[1])
    result = 'Program error: LogDrive("' + cur_dir[1] + '").'
  endif

  if      result == ''
  and not ChDir(cur_dir)
    result = 'Program error: LogDrive("' + cur_dir[1] + '").'
  endif

  // End of setting up the tests.



  // Here the actual tests start ...

  if result == ''
    result = test_SearchPath_test(AddTrailingSlash(path2) + unique_name,
                                  path1 + ';' + path2 + ';' + path3,
                                  '')
  endif

  if result == ''
    result = test_SearchPath_test(unique_name,
                                  path1 + ';' + path2 + ';' + path3,
                                  '')
  endif

  if result == ''
    result = test_SearchPath_test(unique_name,
                                  path1 + ';' + path2 + ';' + path3,
                                  subdir)
  endif

  if result == ''
    result = test_SearchPath_test(unique_name,
                                  path1 + ';' + path2 + ';' + path3,
                                  '.')
  endif

  //  The following (default commented) meta-tests intentionally produce
  //  an error, each in a different form.

  //  test_SearchPath cannot test with an empty path parameter.
  if result == ''
    // result = test_SearchPath_test(unique_name, '', subdir)
  endif

  //  test_SearchPath only works for subdir parameter "mac", "." and "".
  if result == ''
    // result = test_SearchPath_test(unique_name, path1 + ';' + path2, 'NotMac')
  endif

  // Here the actual tests end.



  // Leave the test's temporary current directory so it becomes deletable.
  if (WhichOS() in _WINDOWSNT_, _WINDOWS_)
    LogDrive(old_current_dir[1])
  endif
  ChDir(old_current_dir)

  // Delete all helper files.
  test_SearchPath_delete_files(cur_dir  )
  test_SearchPath_delete_files(path1    )
  test_SearchPath_delete_files(path2    )
  test_SearchPath_delete_files(path3    )
  test_SearchPath_delete_files(LoadDir())

  // Delete all helper directories.
  test_SearchPath_delete_dirs(cur_dir)
  test_SearchPath_delete_dirs(path1  )
  test_SearchPath_delete_dirs(path2  )
  test_SearchPath_delete_dirs(path3  )
  RmDir(wrk_dir)

  // Restore location and clean-up temp file.
  PopLocation()
  AbandonFile(test_SearchPath_id)

  if not test_SearchPath_debug
    AbandonFile(test_SearchPath_debug_id)
  endif

  if result <> ''
    result = 'test_SearchPath: ' + result
  endif

  return(result)
end test_SearchPath


// End of test_SearchPath() implementation.





proc Main()
  string result [MAXSTRINGLEN] = ''
  result = test_SearchPath()
  Message('Test_SearchPath returns an empty string if no errors were found.')
  Warn('"', result, '"')
end Main


